@@ -5,8 +5,6 @@
 //! type itself.
 
 use std::fmt;
-#[cfg(feature = "nightly")]
-use std::marker::Unsize;
 use std::mem::{self, ManuallyDrop};
 use std::prelude::v1::*;
 
@@ -31,9 +29,7 @@ use crate::UnwrapThrowExt;
 ///
 /// The type parameter on `Closure` is the type of closure that this represents.
 /// Currently this can only be the `Fn` and `FnMut` traits with up to 7
-/// arguments (and an optional return value). The arguments/return value of the
-/// trait must be numbers like `u32` for now, although this restriction may be
-/// lifted in the future!
+/// arguments (and an optional return value).
 ///
 /// # Examples
 ///
@@ -48,7 +44,7 @@ use crate::UnwrapThrowExt;
 ///
 /// #[wasm_bindgen]
 /// extern "C" {
-///     fn setInterval(closure: &Closure<FnMut()>, time: u32) -> i32;
+///     fn setInterval(closure: &Closure<dyn FnMut()>, time: u32) -> i32;
 ///     fn clearInterval(id: i32);
 ///
 ///     #[wasm_bindgen(js_namespace = console)]
@@ -58,7 +54,7 @@ use crate::UnwrapThrowExt;
 /// #[wasm_bindgen]
 /// pub struct IntervalHandle {
 ///     interval_id: i32,
-///     _closure: Closure<FnMut()>,
+///     _closure: Closure<dyn FnMut()>,
 /// }
 ///
 /// impl Drop for IntervalHandle {
@@ -69,11 +65,11 @@ use crate::UnwrapThrowExt;
 ///
 /// #[wasm_bindgen]
 /// pub fn run() -> IntervalHandle {
-///     // First up we use `Closure::wrap` to wrap up a Rust closure and create
+///     // First up we use `Closure::new` to wrap up a Rust closure and create
 ///     // a JS closure.
-///     let cb = Closure::wrap(Box::new(|| {
+///     let cb = Closure::new(|| {
 ///         log("interval elapsed!");
-///     }) as Box<FnMut()>);
+///     });
 ///
 ///     // Next we pass this via reference to the `setInterval` function, and
 ///     // `setInterval` gets a handle to the corresponding JS closure.
@@ -101,7 +97,7 @@ use crate::UnwrapThrowExt;
 /// #[wasm_bindgen]
 /// pub struct IntervalHandle {
 ///     interval_id: i32,
-///     _closure: Closure<FnMut()>,
+///     _closure: Closure<dyn FnMut()>,
 /// }
 ///
 /// impl Drop for IntervalHandle {
@@ -113,9 +109,9 @@ use crate::UnwrapThrowExt;
 ///
 /// #[wasm_bindgen]
 /// pub fn run() -> Result<IntervalHandle, JsValue> {
-///     let cb = Closure::wrap(Box::new(|| {
-///         web_sys::console::log_1(&"inverval elapsed!".into());
-///     }) as Box<FnMut()>);
+///     let cb = Closure::new(|| {
+///         web_sys::console::log_1(&"interval elapsed!".into());
+///     });
 ///
 ///     let window = web_sys::window().unwrap();
 ///     let interval_id = window.set_interval_with_callback_and_timeout_and_arguments_0(
@@ -144,7 +140,7 @@ use crate::UnwrapThrowExt;
 ///
 /// #[wasm_bindgen]
 /// extern "C" {
-///     fn requestAnimationFrame(closure: &Closure<FnMut()>) -> u32;
+///     fn requestAnimationFrame(closure: &Closure<dyn FnMut()>) -> u32;
 ///     fn cancelAnimationFrame(id: u32);
 ///
 ///     #[wasm_bindgen(js_namespace = console)]
@@ -154,7 +150,7 @@ use crate::UnwrapThrowExt;
 /// #[wasm_bindgen]
 /// pub struct AnimationFrameHandle {
 ///     animation_id: u32,
-///     _closure: Closure<FnMut()>,
+///     _closure: Closure<dyn FnMut()>,
 /// }
 ///
 /// impl Drop for AnimationFrameHandle {
@@ -254,24 +250,9 @@ impl<T> Closure<T>
 where
     T: ?Sized + WasmClosure,
 {
-    /// A more ergonomic version of `Closure::wrap` that does the boxing and
-    /// cast to trait object for you.
+    /// Creates a new instance of `Closure` from the provided Rust function.
     ///
-    /// *This method requires the `nightly` feature of the `wasm-bindgen` crate
-    /// to be enabled, meaning this is a nightly-only API. Users on stable
-    /// should use `Closure::wrap`.*
-    #[cfg(feature = "nightly")]
-    pub fn new<F>(t: F) -> Closure<T>
-    where
-        F: Unsize<T> + 'static,
-    {
-        Closure::wrap(Box::new(t) as Box<T>)
-    }
-
-    /// Creates a new instance of `Closure` from the provided boxed Rust
-    /// function.
-    ///
-    /// Note that the closure provided here, `Box<T>`, has a few requirements
+    /// Note that the closure provided here, `F`, has a few requirements
     /// associated with it:
     ///
     /// * It must implement `Fn` or `FnMut` (for `FnOnce` functions see
@@ -285,6 +266,15 @@ where
     /// * Its arguments and return values are all types that can be shared with
     ///   JS (i.e. have `#[wasm_bindgen]` annotations or are simple numbers,
     ///   etc.)
+    pub fn new<F>(t: F) -> Closure<T>
+    where
+        F: IntoWasmClosure<T> + 'static,
+    {
+        Closure::wrap(Box::new(t).unsize())
+    }
+
+    /// A more direct version of `Closure::new` which creates a `Closure` from
+    /// a `Box<dyn Fn>`/`Box<dyn FnMut>`, which is how it's kept internally.
     pub fn wrap(mut data: Box<T>) -> Closure<T> {
         assert_eq!(mem::size_of::<*const T>(), mem::size_of::<FatPtr<T>>());
         let (a, b) = unsafe {
@@ -356,33 +346,44 @@ where
         }
     }
 
-    /// Leaks this `Closure` to ensure it remains valid for the duration of the
-    /// entire program.
+    /// Release memory management of this closure from Rust to the JS GC.
     ///
-    /// > **Note**: this function will leak memory. It should be used sparingly
-    /// > to ensure the memory leak doesn't affect the program too much.
+    /// When a `Closure` is dropped it will release the Rust memory and
+    /// invalidate the associated JS closure, but this isn't always desired.
+    /// Some callbacks are alive for the entire duration of the program or for a
+    /// lifetime dynamically managed by the JS GC. This function can be used
+    /// to drop this `Closure` while keeping the associated JS function still
+    /// valid.
     ///
-    /// When a `Closure` is dropped it will invalidate the associated JS
-    /// closure, but this isn't always desired. Some callbacks are alive for
-    /// the entire duration of the program, so this can be used to conveniently
-    /// leak this instance of `Closure` while performing as much internal
-    /// cleanup as it can.
+    /// By default this function will leak memory. This can be dangerous if this
+    /// function is called many times in an application because the memory leak
+    /// will overwhelm the page quickly and crash the wasm.
+    ///
+    /// If the browser, however, supports weak references, then this function
+    /// will not leak memory. Instead the Rust memory will be reclaimed when the
+    /// JS closure is GC'd. Weak references are not enabled by default since
+    /// they're still a proposal for the JS standard. They can be enabled with
+    /// `WASM_BINDGEN_WEAKREF=1` when running `wasm-bindgen`, however.
+    pub fn into_js_value(self) -> JsValue {
+        let idx = self.js.idx;
+        mem::forget(self);
+        JsValue::_new(idx)
+    }
+
+    /// Same as `into_js_value`, but doesn't return a value.
     pub fn forget(self) {
-        unsafe {
-            super::__wbindgen_cb_forget(self.js.idx);
-            mem::forget(self);
-        }
+        drop(self.into_js_value());
     }
 }
 
 // NB: we use a specific `T` for this `Closure<T>` impl block to avoid every
 // call site having to provide an explicit, turbo-fished type like
-// `Closure::<FnOnce()>::once(...)`.
+// `Closure::<dyn FnOnce()>::once(...)`.
 impl Closure<dyn FnOnce()> {
     /// Create a `Closure` from a function that can only be called once.
     ///
     /// Since we have no way of enforcing that JS cannot attempt to call this
-    /// `FnOne(A...) -> R` more than once, this produces a `Closure<FnMut(A...)
+    /// `FnOne(A...) -> R` more than once, this produces a `Closure<dyn FnMut(A...)
     /// -> R>` that will dynamically throw a JavaScript error if called more
     /// than once.
     ///
@@ -404,7 +405,7 @@ impl Closure<dyn FnOnce()> {
     ///
     /// // Create a `Closure` from `f`. Note that the `Closure`'s type parameter
     /// // is `FnMut`, even though `f` is `FnOnce`.
-    /// let closure: Closure<FnMut() -> String> = Closure::once(f);
+    /// let closure: Closure<dyn FnMut() -> String> = Closure::once(f);
     /// ```
     pub fn once<F, A, R>(fn_once: F) -> Closure<F::FnMut>
     where
@@ -425,7 +426,6 @@ impl Closure<dyn FnOnce()> {
     /// and everything it closes over will leak.
     ///
     /// ```rust,ignore
-    /// use js_sys;
     /// use wasm_bindgen::{prelude::*, JsCast};
     ///
     /// let f = Closure::once_into_js(move || {
@@ -464,7 +464,7 @@ where
     T: WasmClosure + ?Sized,
 {
     fn describe() {
-        inform(ANYREF);
+        inform(EXTERNREF);
     }
 }
 
@@ -480,6 +480,15 @@ where
     }
 }
 
+impl<'a, T> OptionIntoWasmAbi for &'a Closure<T>
+where
+    T: WasmClosure + ?Sized,
+{
+    fn none() -> Self::Abi {
+        0
+    }
+}
+
 fn _check() {
     fn _assert<T: IntoWasmAbi>() {}
     _assert::<&Closure<dyn Fn()>>();
@@ -523,6 +532,15 @@ pub unsafe trait WasmClosure {
     fn describe();
 }
 
+/// An internal trait for the `Closure` type.
+///
+/// This trait is not stable and it's not recommended to use this in bounds or
+/// implement yourself.
+#[doc(hidden)]
+pub trait IntoWasmClosure<T: ?Sized> {
+    fn unsize(self: Box<Self>) -> Box<T>;
+}
+
 // The memory safety here in these implementations below is a bit tricky. We
 // want to be able to drop the `Closure` object from within the invocation of a
 // `Closure` for cases like promises. That means that while it's running we
@@ -691,6 +709,22 @@ macro_rules! doit {
                 js_val
             }
         }
+
+        impl<T, $($var,)* R> IntoWasmClosure<dyn FnMut($($var),*) -> R> for T
+            where T: 'static + FnMut($($var),*) -> R,
+                  $($var: FromWasmAbi + 'static,)*
+                  R: ReturnWasmAbi + 'static,
+        {
+            fn unsize(self: Box<Self>) -> Box<dyn FnMut($($var),*) -> R> { self }
+        }
+
+        impl<T, $($var,)* R> IntoWasmClosure<dyn Fn($($var),*) -> R> for T
+            where T: 'static + Fn($($var),*) -> R,
+                  $($var: FromWasmAbi + 'static,)*
+                  R: ReturnWasmAbi + 'static,
+        {
+            fn unsize(self: Box<Self>) -> Box<dyn Fn($($var),*) -> R> { self }
+        }
     )*)
 }
 
@@ -713,8 +747,9 @@ doit! {
 // duplicate.
 
 unsafe impl<A, R> WasmClosure for dyn Fn(&A) -> R
-    where A: RefFromWasmAbi,
-          R: ReturnWasmAbi + 'static,
+where
+    A: RefFromWasmAbi,
+    R: ReturnWasmAbi + 'static,
 {
     fn describe() {
         #[allow(non_snake_case)]
@@ -730,8 +765,7 @@ unsafe impl<A, R> WasmClosure for dyn Fn(&A) -> R
             // convert `ret` as it may throw (for `Result`, for
             // example)
             let ret = {
-                let f: *const dyn Fn(&A) -> R =
-                    FatPtr { fields: (a, b) }.ptr;
+                let f: *const dyn Fn(&A) -> R = FatPtr { fields: (a, b) }.ptr;
                 let arg = <A as RefFromWasmAbi>::ref_from_abi(arg);
                 (*f)(&*arg)
             };
@@ -740,17 +774,14 @@ unsafe impl<A, R> WasmClosure for dyn Fn(&A) -> R
 
         inform(invoke::<A, R> as u32);
 
-        unsafe extern fn destroy<A: RefFromWasmAbi, R: ReturnWasmAbi>(
-            a: usize,
-            b: usize,
-        ) {
+        unsafe extern "C" fn destroy<A: RefFromWasmAbi, R: ReturnWasmAbi>(a: usize, b: usize) {
             // See `Fn()` above for why we simply return
             if a == 0 {
                 return;
             }
-            drop(Box::from_raw(FatPtr::<dyn Fn(&A) -> R> {
-                fields: (a, b)
-            }.ptr));
+            drop(Box::from_raw(
+                FatPtr::<dyn Fn(&A) -> R> { fields: (a, b) }.ptr,
+            ));
         }
         inform(destroy::<A, R> as u32);
 
@@ -759,8 +790,9 @@ unsafe impl<A, R> WasmClosure for dyn Fn(&A) -> R
 }
 
 unsafe impl<A, R> WasmClosure for dyn FnMut(&A) -> R
-    where A: RefFromWasmAbi,
-          R: ReturnWasmAbi + 'static,
+where
+    A: RefFromWasmAbi,
+    R: ReturnWasmAbi + 'static,
 {
     fn describe() {
         #[allow(non_snake_case)]
@@ -776,8 +808,7 @@ unsafe impl<A, R> WasmClosure for dyn FnMut(&A) -> R
             // convert `ret` as it may throw (for `Result`, for
             // example)
             let ret = {
-                let f: *const dyn FnMut(&A) -> R =
-                    FatPtr { fields: (a, b) }.ptr;
+                let f: *const dyn FnMut(&A) -> R = FatPtr { fields: (a, b) }.ptr;
                 let f = f as *mut dyn FnMut(&A) -> R;
                 let arg = <A as RefFromWasmAbi>::ref_from_abi(arg);
                 (*f)(&*arg)
@@ -787,17 +818,14 @@ unsafe impl<A, R> WasmClosure for dyn FnMut(&A) -> R
 
         inform(invoke::<A, R> as u32);
 
-        unsafe extern fn destroy<A: RefFromWasmAbi, R: ReturnWasmAbi>(
-            a: usize,
-            b: usize,
-        ) {
+        unsafe extern "C" fn destroy<A: RefFromWasmAbi, R: ReturnWasmAbi>(a: usize, b: usize) {
             // See `Fn()` above for why we simply return
             if a == 0 {
                 return;
             }
-            drop(Box::from_raw(FatPtr::<dyn FnMut(&A) -> R> {
-                fields: (a, b)
-            }.ptr));
+            drop(Box::from_raw(
+                FatPtr::<dyn FnMut(&A) -> R> { fields: (a, b) }.ptr,
+            ));
         }
         inform(destroy::<A, R> as u32);
 
@@ -807,9 +835,10 @@ unsafe impl<A, R> WasmClosure for dyn FnMut(&A) -> R
 
 #[allow(non_snake_case)]
 impl<T, A, R> WasmClosureFnOnce<(&A,), R> for T
-    where T: 'static + FnOnce(&A) -> R,
-          A: RefFromWasmAbi + 'static,
-          R: ReturnWasmAbi + 'static
+where
+    T: 'static + FnOnce(&A) -> R,
+    A: RefFromWasmAbi + 'static,
+    R: ReturnWasmAbi + 'static,
 {
     type FnMut = dyn FnMut(&A) -> R;
 
@@ -822,8 +851,8 @@ impl<T, A, R> WasmClosureFnOnce<(&A,), R> for T
     }
 
     fn into_js_function(self) -> JsValue {
-        use std::rc::Rc;
         use crate::__rt::WasmRefCell;
+        use std::rc::Rc;
 
         let mut me = Some(self);
 
