@@ -3,7 +3,7 @@
 
 #![doc(hidden)]
 
-use crate::{Clamped, JsValue};
+use crate::{Clamped, JsError, JsValue};
 use cfg_if::cfg_if;
 
 macro_rules! tys {
@@ -36,16 +36,18 @@ tys! {
     REFMUT
     SLICE
     VECTOR
-    ANYREF
+    EXTERNREF
+    NAMED_EXTERNREF
     ENUM
     RUST_STRUCT
     CHAR
     OPTIONAL
+    RESULT
     UNIT
     CLAMPED
 }
 
-#[inline(always)] // see `interpret.rs` in the the cli-support crate
+#[inline(always)] // see the wasm-interpreter crate
 pub fn inform(a: u32) {
     unsafe { super::__wbindgen_describe(a) }
 }
@@ -77,7 +79,7 @@ simple! {
     f64 => F64
     bool => BOOLEAN
     char => CHAR
-    JsValue => ANYREF
+    JsValue => EXTERNREF
 }
 
 cfg_if! {
@@ -169,11 +171,10 @@ impl WasmDescribe for () {
     }
 }
 
-// Note that this is only for `ReturnWasmAbi for Result<T, JsValue>`, which
-// throws the result, so we only need to inform about the `T`.
-impl<T: WasmDescribe> WasmDescribe for Result<T, JsValue> {
+impl<T: WasmDescribe, E: Into<JsValue>> WasmDescribe for Result<T, E> {
     fn describe() {
-        T::describe()
+        inform(RESULT);
+        T::describe();
     }
 }
 
@@ -183,3 +184,9 @@ impl<T: WasmDescribe> WasmDescribe for Clamped<T> {
         T::describe();
     }
 }
+
+impl WasmDescribe for JsError {
+    fn describe() {
+        JsValue::describe();
+    }
+}
