@@ -6,15 +6,19 @@
 //! interface.
 
 #![no_std]
+#![allow(coherence_leak_check)]
 #![doc(html_root_url = "https://docs.rs/wasm-bindgen/0.2")]
-#![cfg_attr(feature = "nightly", feature(unsize))]
 
+use core::convert::TryFrom;
 use core::fmt;
 use core::marker;
 use core::mem;
-use core::ops::{Deref, DerefMut};
+use core::ops::{
+    Add, BitAnd, BitOr, BitXor, Deref, DerefMut, Div, Mul, Neg, Not, Rem, Shl, Shr, Sub,
+};
+use core::u32;
 
-use crate::convert::{FromWasmAbi, WasmOptionalF64, WasmSlice};
+use crate::convert::{FromWasmAbi, WasmSlice};
 
 macro_rules! if_std {
     ($($i:item)*) => ($(
@@ -55,19 +59,22 @@ pub mod prelude {
     if_std! {
         pub use crate::closure::Closure;
     }
+
+    pub use crate::JsError;
 }
 
 pub mod convert;
 pub mod describe;
 
 mod cast;
-pub use crate::cast::JsCast;
+pub use crate::cast::{JsCast, JsObject};
+use convert::WasmOption;
 
 if_std! {
     extern crate std;
     use std::prelude::v1::*;
     pub mod closure;
-    mod anyref;
+    mod externref;
 
     mod cache;
     pub use cache::intern::{intern, unintern};
@@ -142,6 +149,15 @@ impl JsValue {
         unsafe { JsValue::_new(__wbindgen_number_new(n)) }
     }
 
+    /// Creates a new JS value which is a bigint from a string representing a number.
+    ///
+    /// This function creates a JS value representing a bigint (a heap
+    /// allocated large integer) and returns a handle to the JS version of it.
+    #[inline]
+    pub fn bigint_from_str(s: &str) -> JsValue {
+        unsafe { JsValue::_new(__wbindgen_bigint_from_str(s.as_ptr(), s.len())) }
+    }
+
     /// Creates a new JS value which is a boolean.
     ///
     /// This function creates a JS object representing a boolean (a heap
@@ -186,6 +202,14 @@ impl JsValue {
     /// Creates a new `JsValue` from the JSON serialization of the object `t`
     /// provided.
     ///
+    /// **This function is deprecated**, due to [creating a dependency cycle in
+    /// some circumstances][dep-cycle-issue]. Use [`serde-wasm-bindgen`] or
+    /// [`gloo_utils::format::JsValueSerdeExt`] instead.
+    ///
+    /// [dep-cycle-issue]: https://github.com/rustwasm/wasm-bindgen/issues/2770
+    /// [`serde-wasm-bindgen`]: https://docs.rs/serde-wasm-bindgen
+    /// [`gloo_utils::format::JsValueSerdeExt`]: https://docs.rs/gloo-utils/latest/gloo_utils/format/trait.JsValueSerdeExt.html
+    ///
     /// This function will serialize the provided value `t` to a JSON string,
     /// send the JSON string to JS, parse it into a JS object, and then return
     /// a handle to the JS object. This is unlikely to be super speedy so it's
@@ -199,6 +223,7 @@ impl JsValue {
     ///
     /// Returns any error encountered when serializing `T` into JSON.
     #[cfg(feature = "serde-serialize")]
+    #[deprecated = "causes dependency cycles, use `serde-wasm-bindgen` or `gloo_utils::format::JsValueSerdeExt` instead"]
     pub fn from_serde<T>(t: &T) -> serde_json::Result<JsValue>
     where
         T: serde::ser::Serialize + ?Sized,
@@ -210,6 +235,14 @@ impl JsValue {
     /// Invokes `JSON.stringify` on this value and then parses the resulting
     /// JSON into an arbitrary Rust value.
     ///
+    /// **This function is deprecated**, due to [creating a dependency cycle in
+    /// some circumstances][dep-cycle-issue]. Use [`serde-wasm-bindgen`] or
+    /// [`gloo_utils::format::JsValueSerdeExt`] instead.
+    ///
+    /// [dep-cycle-issue]: https://github.com/rustwasm/wasm-bindgen/issues/2770
+    /// [`serde-wasm-bindgen`]: https://docs.rs/serde-wasm-bindgen
+    /// [`gloo_utils::format::JsValueSerdeExt`]: https://docs.rs/gloo-utils/latest/gloo_utils/format/trait.JsValueSerdeExt.html
+    ///
     /// This function will first call `JSON.stringify` on the `JsValue` itself.
     /// The resulting string is then passed into Rust which then parses it as
     /// JSON into the resulting value.
@@ -221,6 +254,7 @@ impl JsValue {
     ///
     /// Returns any error encountered when parsing the JSON into a `T`.
     #[cfg(feature = "serde-serialize")]
+    #[deprecated = "causes dependency cycles, use `serde-wasm-bindgen` or `gloo_utils::format::JsValueSerdeExt` instead"]
     pub fn into_serde<T>(&self) -> serde_json::Result<T>
     where
         T: for<'a> serde::de::Deserialize<'a>,
@@ -237,11 +271,13 @@ impl JsValue {
     ///
     /// If this JS value is not an instance of a number then this returns
     /// `None`.
+    #[inline]
     pub fn as_f64(&self) -> Option<f64> {
         unsafe { FromWasmAbi::from_abi(__wbindgen_number_get(self.idx)) }
     }
 
     /// Tests whether this JS value is a JS string.
+    #[inline]
     pub fn is_string(&self) -> bool {
         unsafe { __wbindgen_is_string(self.idx) == 1 }
     }
@@ -267,6 +303,7 @@ impl JsValue {
     ///
     /// [caveats]: https://rustwasm.github.io/docs/wasm-bindgen/reference/types/str.html
     #[cfg(feature = "std")]
+    #[inline]
     pub fn as_string(&self) -> Option<String> {
         unsafe { FromWasmAbi::from_abi(__wbindgen_string_get(self.idx)) }
     }
@@ -276,6 +313,7 @@ impl JsValue {
     ///
     /// If this JS value is not an instance of a boolean then this returns
     /// `None`.
+    #[inline]
     pub fn as_bool(&self) -> Option<bool> {
         unsafe {
             match __wbindgen_boolean_get(self.idx) {
@@ -316,6 +354,28 @@ impl JsValue {
         unsafe { __wbindgen_is_function(self.idx) == 1 }
     }
 
+    /// Tests whether the type of this JS value is `bigint`.
+    #[inline]
+    pub fn is_bigint(&self) -> bool {
+        unsafe { __wbindgen_is_bigint(self.idx) == 1 }
+    }
+
+    /// Applies the unary `typeof` JS operator on a `JsValue`.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/typeof)
+    #[inline]
+    pub fn js_typeof(&self) -> JsValue {
+        unsafe { JsValue::_new(__wbindgen_typeof(self.idx)) }
+    }
+
+    /// Applies the binary `in` JS operator on the two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/in)
+    #[inline]
+    pub fn js_in(&self, obj: &JsValue) -> bool {
+        unsafe { __wbindgen_in(self.idx, obj.idx) == 1 }
+    }
+
     /// Tests whether the value is ["truthy"].
     ///
     /// ["truthy"]: https://developer.mozilla.org/en-US/docs/Glossary/Truthy
@@ -332,7 +392,7 @@ impl JsValue {
         unsafe { __wbindgen_is_falsy(self.idx) == 1 }
     }
 
-    /// Get a string representation of the JavaScript object for debugging
+    /// Get a string representation of the JavaScript object for debugging.
     #[cfg(feature = "std")]
     fn as_debug_string(&self) -> String {
         unsafe {
@@ -342,11 +402,94 @@ impl JsValue {
             String::from_utf8_unchecked(data)
         }
     }
+
+    /// Compare two `JsValue`s for equality, using the `==` operator in JS.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Equality)
+    #[inline]
+    pub fn loose_eq(&self, other: &Self) -> bool {
+        unsafe { __wbindgen_jsval_loose_eq(self.idx, other.idx) != 0 }
+    }
+
+    /// Applies the unary `~` JS operator on a `JsValue`.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Bitwise_NOT)
+    #[inline]
+    pub fn bit_not(&self) -> JsValue {
+        unsafe { JsValue::_new(__wbindgen_bit_not(self.idx)) }
+    }
+
+    /// Applies the binary `>>>` JS operator on the two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Unsigned_right_shift)
+    #[inline]
+    pub fn unsigned_shr(&self, rhs: &Self) -> u32 {
+        unsafe { __wbindgen_unsigned_shr(self.idx, rhs.idx) }
+    }
+
+    /// Applies the binary `/` JS operator on two `JsValue`s, catching and returning any `RangeError` thrown.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Division)
+    #[inline]
+    pub fn checked_div(&self, rhs: &Self) -> Self {
+        unsafe { JsValue::_new(__wbindgen_checked_div(self.idx, rhs.idx)) }
+    }
+
+    /// Applies the binary `**` JS operator on the two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Exponentiation)
+    #[inline]
+    pub fn pow(&self, rhs: &Self) -> Self {
+        unsafe { JsValue::_new(__wbindgen_pow(self.idx, rhs.idx)) }
+    }
+
+    /// Applies the binary `<` JS operator on the two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Less_than)
+    #[inline]
+    pub fn lt(&self, other: &Self) -> bool {
+        unsafe { __wbindgen_lt(self.idx, other.idx) == 1 }
+    }
+
+    /// Applies the binary `<=` JS operator on the two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Less_than_or_equal)
+    #[inline]
+    pub fn le(&self, other: &Self) -> bool {
+        unsafe { __wbindgen_le(self.idx, other.idx) == 1 }
+    }
+
+    /// Applies the binary `>=` JS operator on the two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Greater_than_or_equal)
+    #[inline]
+    pub fn ge(&self, other: &Self) -> bool {
+        unsafe { __wbindgen_ge(self.idx, other.idx) == 1 }
+    }
+
+    /// Applies the binary `>` JS operator on the two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Greater_than)
+    #[inline]
+    pub fn gt(&self, other: &Self) -> bool {
+        unsafe { __wbindgen_gt(self.idx, other.idx) == 1 }
+    }
+
+    /// Applies the unary `+` JS operator on a `JsValue`. Can throw.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Unary_plus)
+    #[inline]
+    pub fn unchecked_into_f64(&self) -> f64 {
+        unsafe { __wbindgen_as_number(self.idx) }
+    }
 }
 
 impl PartialEq for JsValue {
+    /// Compares two `JsValue`s for equality, using the `===` operator in JS.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Strict_equality)
     #[inline]
-    fn eq(&self, other: &JsValue) -> bool {
+    fn eq(&self, other: &Self) -> bool {
         unsafe { __wbindgen_jsval_eq(self.idx, other.idx) != 0 }
     }
 }
@@ -387,6 +530,248 @@ if_std! {
     }
 }
 
+macro_rules! forward_deref_unop {
+    (impl $imp:ident, $method:ident for $t:ty) => {
+        impl $imp for $t {
+            type Output = <&'static $t as $imp>::Output;
+
+            #[inline]
+            fn $method(self) -> <&'static $t as $imp>::Output {
+                $imp::$method(&self)
+            }
+        }
+    };
+}
+
+macro_rules! forward_deref_binop {
+    (impl $imp:ident, $method:ident for $t:ty) => {
+        impl<'a> $imp<$t> for &'a $t {
+            type Output = <&'static $t as $imp<&'static $t>>::Output;
+
+            #[inline]
+            fn $method(self, other: $t) -> <&'static $t as $imp<&'static $t>>::Output {
+                $imp::$method(self, &other)
+            }
+        }
+
+        impl $imp<&$t> for $t {
+            type Output = <&'static $t as $imp<&'static $t>>::Output;
+
+            #[inline]
+            fn $method(self, other: &$t) -> <&'static $t as $imp<&'static $t>>::Output {
+                $imp::$method(&self, other)
+            }
+        }
+
+        impl $imp<$t> for $t {
+            type Output = <&'static $t as $imp<&'static $t>>::Output;
+
+            #[inline]
+            fn $method(self, other: $t) -> <&'static $t as $imp<&'static $t>>::Output {
+                $imp::$method(&self, &other)
+            }
+        }
+    };
+}
+
+impl Not for &JsValue {
+    type Output = bool;
+
+    /// Applies the `!` JS operator on a `JsValue`.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Logical_NOT)
+    #[inline]
+    fn not(self) -> Self::Output {
+        JsValue::is_falsy(self)
+    }
+}
+
+forward_deref_unop!(impl Not, not for JsValue);
+
+impl TryFrom<JsValue> for f64 {
+    type Error = JsValue;
+
+    /// Applies the unary `+` JS operator on a `JsValue`.
+    /// Returns the numeric result on success, or the JS error value on error.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Unary_plus)
+    #[inline]
+    fn try_from(val: JsValue) -> Result<Self, Self::Error> {
+        f64::try_from(&val)
+    }
+}
+
+impl TryFrom<&JsValue> for f64 {
+    type Error = JsValue;
+
+    /// Applies the unary `+` JS operator on a `JsValue`.
+    /// Returns the numeric result on success, or the JS error value on error.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Unary_plus)
+    #[inline]
+    fn try_from(val: &JsValue) -> Result<Self, Self::Error> {
+        let jsval = unsafe { JsValue::_new(__wbindgen_try_into_number(val.idx)) };
+        return match jsval.as_f64() {
+            Some(num) => Ok(num),
+            None => Err(jsval),
+        };
+    }
+}
+
+impl Neg for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the unary `-` JS operator on a `JsValue`.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Unary_negation)
+    #[inline]
+    fn neg(self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_neg(self.idx)) }
+    }
+}
+
+forward_deref_unop!(impl Neg, neg for JsValue);
+
+impl BitAnd for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `&` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Bitwise_AND)
+    #[inline]
+    fn bitand(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_bit_and(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl BitAnd, bitand for JsValue);
+
+impl BitOr for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `|` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Bitwise_OR)
+    #[inline]
+    fn bitor(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_bit_or(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl BitOr, bitor for JsValue);
+
+impl BitXor for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `^` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Bitwise_XOR)
+    #[inline]
+    fn bitxor(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_bit_xor(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl BitXor, bitxor for JsValue);
+
+impl Shl for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `<<` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Left_shift)
+    #[inline]
+    fn shl(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_shl(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl Shl, shl for JsValue);
+
+impl Shr for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `>>` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Right_shift)
+    #[inline]
+    fn shr(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_shr(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl Shr, shr for JsValue);
+
+impl Add for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `+` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Addition)
+    #[inline]
+    fn add(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_add(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl Add, add for JsValue);
+
+impl Sub for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `-` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Subtraction)
+    #[inline]
+    fn sub(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_sub(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl Sub, sub for JsValue);
+
+impl Div for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `/` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Division)
+    #[inline]
+    fn div(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_div(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl Div, div for JsValue);
+
+impl Mul for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `*` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Multiplication)
+    #[inline]
+    fn mul(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_mul(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl Mul, mul for JsValue);
+
+impl Rem for &JsValue {
+    type Output = JsValue;
+
+    /// Applies the binary `%` JS operator on two `JsValue`s.
+    ///
+    /// [MDN documentation](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Remainder)
+    #[inline]
+    fn rem(self, rhs: Self) -> Self::Output {
+        unsafe { JsValue::_new(__wbindgen_rem(self.idx, rhs.idx)) }
+    }
+}
+
+forward_deref_binop!(impl Rem, rem for JsValue);
+
 impl<'a> From<&'a str> for JsValue {
     #[inline]
     fn from(s: &'a str) -> JsValue {
@@ -483,6 +868,119 @@ macro_rules! numbers {
 
 numbers! { i8 u8 i16 u16 i32 u32 f32 f64 }
 
+macro_rules! big_numbers {
+    (|$arg:ident|, $($n:ident = $handle:expr,)*) => ($(
+        impl PartialEq<$n> for JsValue {
+            #[inline]
+            fn eq(&self, other: &$n) -> bool {
+                self == &JsValue::from(*other)
+            }
+        }
+
+        impl From<$n> for JsValue {
+            #[inline]
+            fn from($arg: $n) -> JsValue {
+                unsafe { JsValue::_new($handle) }
+            }
+        }
+    )*)
+}
+
+fn bigint_get_as_i64(v: &JsValue) -> Option<i64> {
+    unsafe { Option::from_abi(__wbindgen_bigint_get_as_i64(v.idx)) }
+}
+
+macro_rules! try_from_for_num64 {
+    ($ty:ty) => {
+        impl TryFrom<JsValue> for $ty {
+            type Error = JsValue;
+
+            #[inline]
+            fn try_from(v: JsValue) -> Result<Self, JsValue> {
+                bigint_get_as_i64(&v)
+                    // Reinterpret bits; ABI-wise this is safe to do and allows us to avoid
+                    // having separate intrinsics per signed/unsigned types.
+                    .map(|as_i64| as_i64 as Self)
+                    // Double-check that we didn't truncate the bigint to 64 bits.
+                    .filter(|as_self| v == *as_self)
+                    // Not a bigint or not in range.
+                    .ok_or(v)
+            }
+        }
+    };
+}
+
+try_from_for_num64!(i64);
+try_from_for_num64!(u64);
+
+macro_rules! try_from_for_num128 {
+    ($ty:ty, $hi_ty:ty) => {
+        impl TryFrom<JsValue> for $ty {
+            type Error = JsValue;
+
+            #[inline]
+            fn try_from(v: JsValue) -> Result<Self, JsValue> {
+                // Truncate the bigint to 64 bits, this will give us the lower part.
+                let lo = match bigint_get_as_i64(&v) {
+                    // The lower part must be interpreted as unsigned in both i128 and u128.
+                    Some(lo) => lo as u64,
+                    // Not a bigint.
+                    None => return Err(v),
+                };
+                // Now we know it's a bigint, so we can safely use `>> 64n` without
+                // worrying about a JS exception on type mismatch.
+                let hi = v >> JsValue::from(64_u64);
+                // The high part is the one we want checked against a 64-bit range.
+                // If it fits, then our original number is in the 128-bit range.
+                let hi = <$hi_ty>::try_from(hi)?;
+                Ok(Self::from(hi) << 64 | Self::from(lo))
+            }
+        }
+    };
+}
+
+try_from_for_num128!(i128, i64);
+try_from_for_num128!(u128, u64);
+
+big_numbers! {
+    |n|,
+    i64 = __wbindgen_bigint_from_i64(n),
+    u64 = __wbindgen_bigint_from_u64(n),
+    i128 = __wbindgen_bigint_from_i128((n >> 64) as i64, n as u64),
+    u128 = __wbindgen_bigint_from_u128((n >> 64) as u64, n as u64),
+}
+
+// `usize` and `isize` have to be treated a bit specially, because we know that
+// they're 32-bit but the compiler conservatively assumes they might be bigger.
+// So, we have to manually forward to the `u32`/`i32` versions.
+impl PartialEq<usize> for JsValue {
+    #[inline]
+    fn eq(&self, other: &usize) -> bool {
+        *self == (*other as u32)
+    }
+}
+
+impl From<usize> for JsValue {
+    #[inline]
+    fn from(n: usize) -> Self {
+        Self::from(n as u32)
+    }
+}
+
+impl PartialEq<isize> for JsValue {
+    #[inline]
+    fn eq(&self, other: &isize) -> bool {
+        *self == (*other as i32)
+    }
+}
+
+impl From<isize> for JsValue {
+    #[inline]
+    fn from(n: isize) -> Self {
+        Self::from(n as i32)
+    }
+}
+
 externs! {
     #[link(wasm_import_module = "__wbindgen_placeholder__")]
     extern "C" {
@@ -491,10 +989,15 @@ externs! {
 
         fn __wbindgen_string_new(ptr: *const u8, len: usize) -> u32;
         fn __wbindgen_number_new(f: f64) -> u32;
+        fn __wbindgen_bigint_from_str(ptr: *const u8, len: usize) -> u32;
+        fn __wbindgen_bigint_from_i64(n: i64) -> u32;
+        fn __wbindgen_bigint_from_u64(n: u64) -> u32;
+        fn __wbindgen_bigint_from_i128(hi: i64, lo: u64) -> u32;
+        fn __wbindgen_bigint_from_u128(hi: u64, lo: u64) -> u32;
         fn __wbindgen_symbol_named_new(ptr: *const u8, len: usize) -> u32;
         fn __wbindgen_symbol_anonymous_new() -> u32;
 
-        fn __wbindgen_anyref_heap_live_count() -> u32;
+        fn __wbindgen_externref_heap_live_count() -> u32;
 
         fn __wbindgen_is_null(idx: u32) -> u32;
         fn __wbindgen_is_undefined(idx: u32) -> u32;
@@ -502,19 +1005,46 @@ externs! {
         fn __wbindgen_is_object(idx: u32) -> u32;
         fn __wbindgen_is_function(idx: u32) -> u32;
         fn __wbindgen_is_string(idx: u32) -> u32;
-        fn __wbindgen_is_falsy(idx: u32) -> u32;
+        fn __wbindgen_is_bigint(idx: u32) -> u32;
+        fn __wbindgen_typeof(idx: u32) -> u32;
 
-        fn __wbindgen_number_get(idx: u32) -> WasmOptionalF64;
+        fn __wbindgen_in(prop: u32, obj: u32) -> u32;
+
+        fn __wbindgen_is_falsy(idx: u32) -> u32;
+        fn __wbindgen_as_number(idx: u32) -> f64;
+        fn __wbindgen_try_into_number(idx: u32) -> u32;
+        fn __wbindgen_neg(idx: u32) -> u32;
+        fn __wbindgen_bit_and(a: u32, b: u32) -> u32;
+        fn __wbindgen_bit_or(a: u32, b: u32) -> u32;
+        fn __wbindgen_bit_xor(a: u32, b: u32) -> u32;
+        fn __wbindgen_bit_not(idx: u32) -> u32;
+        fn __wbindgen_shl(a: u32, b: u32) -> u32;
+        fn __wbindgen_shr(a: u32, b: u32) -> u32;
+        fn __wbindgen_unsigned_shr(a: u32, b: u32) -> u32;
+        fn __wbindgen_add(a: u32, b: u32) -> u32;
+        fn __wbindgen_sub(a: u32, b: u32) -> u32;
+        fn __wbindgen_div(a: u32, b: u32) -> u32;
+        fn __wbindgen_checked_div(a: u32, b: u32) -> u32;
+        fn __wbindgen_mul(a: u32, b: u32) -> u32;
+        fn __wbindgen_rem(a: u32, b: u32) -> u32;
+        fn __wbindgen_pow(a: u32, b: u32) -> u32;
+        fn __wbindgen_lt(a: u32, b: u32) -> u32;
+        fn __wbindgen_le(a: u32, b: u32) -> u32;
+        fn __wbindgen_ge(a: u32, b: u32) -> u32;
+        fn __wbindgen_gt(a: u32, b: u32) -> u32;
+
+        fn __wbindgen_number_get(idx: u32) -> WasmOption<f64>;
         fn __wbindgen_boolean_get(idx: u32) -> u32;
         fn __wbindgen_string_get(idx: u32) -> WasmSlice;
+        fn __wbindgen_bigint_get_as_i64(idx: u32) -> WasmOption<i64>;
 
         fn __wbindgen_debug_string(ret: *mut [usize; 2], idx: u32) -> ();
 
         fn __wbindgen_throw(a: *const u8, b: usize) -> !;
         fn __wbindgen_rethrow(a: u32) -> !;
+        fn __wbindgen_error_new(a: *const u8, b: usize) -> u32;
 
         fn __wbindgen_cb_drop(idx: u32) -> u32;
-        fn __wbindgen_cb_forget(idx: u32) -> ();
 
         fn __wbindgen_describe(v: u32) -> ();
         fn __wbindgen_describe_closure(a: u32, b: u32, c: u32) -> u32;
@@ -522,6 +1052,9 @@ externs! {
         fn __wbindgen_json_parse(ptr: *const u8, len: usize) -> u32;
         fn __wbindgen_json_serialize(idx: u32) -> WasmSlice;
         fn __wbindgen_jsval_eq(a: u32, b: u32) -> u32;
+        fn __wbindgen_jsval_loose_eq(a: u32, b: u32) -> u32;
+
+        fn __wbindgen_not(idx: u32) -> u32;
 
         fn __wbindgen_memory() -> u32;
         fn __wbindgen_module() -> u32;
@@ -570,6 +1103,12 @@ impl Drop for JsValue {
     }
 }
 
+impl Default for JsValue {
+    fn default() -> Self {
+        Self::UNDEFINED
+    }
+}
+
 /// Wrapper type for imported statics.
 ///
 /// This type is used whenever a `static` is imported from a JS module, for
@@ -658,7 +1197,7 @@ pub fn throw_val(s: JsValue) -> ! {
     }
 }
 
-/// Get the count of live `anyref`s / `JsValue`s in `wasm-bindgen`'s heap.
+/// Get the count of live `externref`s / `JsValue`s in `wasm-bindgen`'s heap.
 ///
 /// ## Usage
 ///
@@ -670,7 +1209,7 @@ pub fn throw_val(s: JsValue) -> ! {
 /// * get an initial live count,
 ///
 /// * perform some series of operations or function calls that should clean up
-///   after themselves, and should not keep holding onto `anyref`s / `JsValue`s
+///   after themselves, and should not keep holding onto `externref`s / `JsValue`s
 ///   after completion,
 ///
 /// * get the final live count,
@@ -679,8 +1218,8 @@ pub fn throw_val(s: JsValue) -> ! {
 ///
 /// ## What is Counted
 ///
-/// Note that this only counts the *owned* `anyref`s / `JsValue`s that end up in
-/// `wasm-bindgen`'s heap. It does not count borrowed `anyref`s / `JsValue`s
+/// Note that this only counts the *owned* `externref`s / `JsValue`s that end up in
+/// `wasm-bindgen`'s heap. It does not count borrowed `externref`s / `JsValue`s
 /// that are on its stack.
 ///
 /// For example, these `JsValue`s are accounted for:
@@ -689,7 +1228,7 @@ pub fn throw_val(s: JsValue) -> ! {
 /// #[wasm_bindgen]
 /// pub fn my_function(this_is_counted: JsValue) {
 ///     let also_counted = JsValue::from_str("hi");
-///     assert!(wasm_bindgen::anyref_heap_live_count() >= 2);
+///     assert!(wasm_bindgen::externref_heap_live_count() >= 2);
 /// }
 /// ```
 ///
@@ -702,11 +1241,16 @@ pub fn throw_val(s: JsValue) -> ! {
 ///     // ...
 /// }
 /// ```
-pub fn anyref_heap_live_count() -> u32 {
-    unsafe { __wbindgen_anyref_heap_live_count() }
+pub fn externref_heap_live_count() -> u32 {
+    unsafe { __wbindgen_externref_heap_live_count() }
 }
 
-/// An extension trait for `Option<T>` and `Result<T, E>` for unwraping the `T`
+#[doc(hidden)]
+pub fn anyref_heap_live_count() -> u32 {
+    externref_heap_live_count()
+}
+
+/// An extension trait for `Option<T>` and `Result<T, E>` for unwrapping the `T`
 /// value, or throwing a JS error if it is not available.
 ///
 /// These methods should have a smaller code size footprint than the normal
@@ -1054,7 +1598,7 @@ pub mod __rt {
     ///
     /// Ideas for how to improve this are most welcome!
     pub fn link_mem_intrinsics() {
-        crate::anyref::link_intrinsics();
+        crate::externref::link_intrinsics();
     }
 
     static mut GLOBAL_EXNDATA: [u32; 2] = [0; 2];
@@ -1166,3 +1710,91 @@ impl<T> DerefMut for Clamped<T> {
         &mut self.0
     }
 }
+
+/// Convenience type for use on exported `fn() -> Result<T, JsError>` functions, where you wish to
+/// throw a JavaScript `Error` object.
+///
+/// You can get wasm_bindgen to throw basic errors by simply returning
+/// `Err(JsError::new("message"))` from such a function.
+///
+/// For more complex error handling, `JsError` implements `From<T> where T: std::error::Error` by
+/// converting it to a string, so you can use it with `?`. Many Rust error types already do this,
+/// and you can use [`thiserror`](https://crates.io/crates/thiserror) to derive Display
+/// implementations easily or use any number of boxed error types that implement it already.
+///
+///
+/// To allow JavaScript code to catch only your errors, you may wish to add a subclass of `Error`
+/// in a JS module, and then implement `Into<JsValue>` directly on a type and instantiate that
+/// subclass. In that case, you would not need `JsError` at all.
+///
+/// ### Basic example
+///
+/// ```rust,no_run
+/// use wasm_bindgen::prelude::*;
+///
+/// #[wasm_bindgen]
+/// pub fn throwing_function() -> Result<(), JsError> {
+///     Err(JsError::new("message"))
+/// }
+/// ```
+///
+/// ### Complex Example
+///
+/// ```rust,no_run
+/// use wasm_bindgen::prelude::*;
+///
+/// #[derive(Debug, Clone)]
+/// enum MyErrorType {
+///     SomeError,
+/// }
+///
+/// use core::fmt;
+/// impl std::error::Error for MyErrorType {}
+/// impl fmt::Display for MyErrorType {
+///     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+///         write!(f, "display implementation becomes the error message")
+///     }
+/// }
+///
+/// fn internal_api() -> Result<(), MyErrorType> {
+///     Err(MyErrorType::SomeError)
+/// }
+///
+/// #[wasm_bindgen]
+/// pub fn throwing_function() -> Result<(), JsError> {
+///     internal_api()?;
+///     Ok(())
+/// }
+///
+/// ```
+#[derive(Clone)]
+pub struct JsError {
+    value: JsValue,
+}
+
+impl JsError {
+    /// Construct a JavaScript `Error` object with a string message
+    #[inline]
+    pub fn new(s: &str) -> JsError {
+        Self {
+            value: unsafe { JsValue::_new(crate::__wbindgen_error_new(s.as_ptr(), s.len())) },
+        }
+    }
+}
+
+if_std! {
+    impl<E> From<E> for JsError
+    where
+        E: std::error::Error,
+    {
+        fn from(error: E) -> Self {
+            JsError::new(&error.to_string())
+        }
+    }
+}
+
+impl From<JsError> for JsValue {
+    fn from(error: JsError) -> Self {
+        error.value
+    }
+}
